package cn.ischoolcode.javabook.dushichat;

import java.io.*;
import java.util.*;

/**
 * 运行在服务器端的文件传输线程。它接受
 * 从客户端发送传入的文件块，并把它们发送到以协调的方式接收
 * 的客户端，直到所有的用户要么拒绝或接收
 * 文件。如果没有接收到更多的用户和到期时间已
 * 达到了该线程将终止。 
 * 
 * @author zcj
 */
public class DushiServerFileTransfer extends DushiFileTransfer {
	private DushiServerConnection connection = null;
	private Vector invitedUsers = null;
	private Vector transferUsers = null;
	private boolean sleeping = false;

	public DushiServerFileTransfer(DushiServerConnection c, String fn, long ts,
			int tb, int bs, int minutes, Vector iu) {
		super(fn, ts, tb, bs, minutes);

		connection = c;
		invitedUsers = iu;
		transferUsers = new Vector(invitedUsers.size());

		try {
			tmpFile = File.createTempFile(fileName, null);
			transFile = new RandomAccessFile(tmpFile, "rw");
		} catch (Exception e) {
			return;
		}

		start();
	}

	/**
	 * 循环通过所有传输用户，将文件块传输给他们。当所有文件块都已传输给用户时，
	 * 它们将从列表中删除。如果用户已断开连接，则会将其从列表中删除。当过期日
	 * 期已经超过并且活动列表中没有更多用户时，线程终止。
	 */
	public void run() {
		while ((transferUsers.size() > 0)
				|| ((invitedUsers.size() > 0) && (System.currentTimeMillis() <= expiryTime))) {
			// 在发送所有文件块之前，请确保发件人没有断开连接
			if (!connection.online && (receivedBlocks < totalBlocks)) {
				// 看起来发件人过早断开了连接，因此我们需要终止文件传输
				cancel();
				break;
			}

			if (transferUsers.size() <= 0) {
				// 等一秒。
				try {
					sleeping = true;
					sleep(1000);
				} catch (InterruptedException e) {
				}
				sleeping = false;
				continue;
			}

			Enumeration e = transferUsers.elements();
			while (e.hasMoreElements()) {
				TransferUser user = (TransferUser) e.nextElement();

				if (!user.connection.online || (user.nextBlock >= totalBlocks)) {
					// 断开连接或已完成。
					transferUsers.remove(user);
					continue;
				}

				if (user.nextBlock >= receivedBlocks)
					// 我们还没有此用户的下一个区块。
					continue;

				try {
					byte[] data = getFileBlock(user.nextBlock);

					// 将下一个文件块发送给用户。
					user.connection.sendFileBlock(connection.user.id, fileName,
							user.nextBlock, data);
				} catch (IOException f) {
				}

				user.nextBlock += 1;
			}
		}

		// 无论如何，我们完成了。
		cancel();

		try {
			transFile.close();
		} catch (IOException e) {
		}

		tmpFile.delete();
	}

	/**
	 * 将已接受邀请的受邀用户添加到参与用户列表中，并开始发送文件块。
	 */
	public void addUser(DushiServerConnection c) {
		// 这个用户在我们的邀请列表中吗？
		if (invitedUsers.remove(c)) {
			transferUsers.add(new TransferUser(c));

			// 如果正在休眠，等待转移用户，请唤醒 。
			if (sleeping)
				interrupt();
		}
	}

	/**
	 * 从参与用户列表中删除用户。
	 */
	public boolean removeUser(DushiServerConnection c) {
		Enumeration e = null;

		e = transferUsers.elements();
		while (e.hasMoreElements()) {
			TransferUser user = (TransferUser) e.nextElement();

			if (user.connection == c) {
				transferUsers.remove(user);
				return (true);
			}
		}
		return (false);
	}

	/**
	 * 终止文件传输。
	 */
	public void cancel() {
		Enumeration e = null;

		e = transferUsers.elements();
		while (e.hasMoreElements()) {
			TransferUser user = (TransferUser) e.nextElement();

			user.connection.sendFileCancel(connection.user.id, fileName);
		}
		transferUsers.removeAllElements();

		e = invitedUsers.elements();
		while (e.hasMoreElements()) {
			DushiServerConnection c = (DushiServerConnection) e.nextElement();

			c.sendFileCancel(connection.user.id, fileName);
		}
		invitedUsers.removeAllElements();
	}

	class TransferUser {
		protected final DushiServerConnection connection;
		protected int nextBlock;

		public TransferUser(DushiServerConnection c) {
			connection = c;
			nextBlock = 0;
		}
	}
}
